from typing import List, Callable, NamedTuple

from pynanto import Response
from pynanto.response import Request


class Route(NamedTuple):
    path: str
    callback: Callable[[Request], Response]


class Routes:
    def __init__(self):
        self.list: List[Route] = []

    def add_route(self, path: str, callback: Callable[[Request], Response]) -> 'Routes':
        self.list.append(Route(path, callback))
        return self

    def add_route_obj(self, route: Route) -> 'Routes':
        self.list.append(route)
        return self

    @property
    def is_empty(self) -> bool:
        return len(self.list) == 0



from os import getcwd

from pynanto import Config


# if main it does not detect the 'user folder' correctly


def quickstart():
    print(getcwd())
    Config().quickstart()


if __name__ == '__main__':
    quickstart()



from typing import NamedTuple, Union


class Response(NamedTuple):
    content: Union[str, bytes]
    content_type: str

    @staticmethod
    def application_zip(content: bytes):
        content_type = 'application/zip, application/octet-stream, application/x-zip-compressed, multipart/x-zip'
        return Response(content, content_type)


class Request(NamedTuple):
    method: str
    content: Union[str, bytes]
    content_type: str



from __future__ import annotations

from typing import TYPE_CHECKING

from .bootstrap import Bootstrap
from .bundle import Bundle
from .bundles import Bundles
from .resource import Resource, StringResource, PathResource
from .response import Response
from .routes import Routes, Route
from .webserver import Webserver
from .webservers.available_webservers import available_webservers

if True:  # workaround to stop Pycharm reordering Config import
    pass

from .config import Config

if TYPE_CHECKING:
    from .webservers.flask import WsFlask
    from .webservers.fastapi import WsFastapi


def ws_flask() -> WsFlask:
    from .webservers.flask import WsFlask
    return WsFlask()


def ws_fastapi() -> WsFastapi:
    from .webservers.fastapi import WsFastapi
    return WsFastapi()


__version__ = '0.0.14'



from pathlib import Path
from typing import List


class Bootstrap:
    def __init__(self):
        self._remote_python: List[str] = []
        self.default_python = _wrap_in_try_except(
            'from js import document\n' +
            'document.body.innerHTML = "<h1>bootstrap done. No remote python specified</h1>"'
        )

    def add_python(self, code: str) -> 'Bootstrap':
        self._remote_python.append(_wrap_in_try_except(code))
        return self

    def get_javascript(self) -> str:
        rp = self._remote_python.copy()
        if len(rp) == 0:
            rp.append(self.default_python)
        with_indent = map(lambda l: '' + l, rp)
        parent = Path(__file__).parent
        # js = parent / 'bootstrap_pyodide.js'
        # js_content = js.read_text()
        python_log = 'print("first instruction run by pyodide")\n'
        return js_content.replace('# python replace marker', python_log + '\n'.join(with_indent))


def _wrap_in_try_except(code: str) -> str:
    import textwrap
    # language=python
    return f"""
try:
    pass
except Exception as ex:
    import traceback
    traceback.print_exc()            
""".replace('pass', '\n' + textwrap.indent(code, ' ' * 4))


js_content = """
if (typeof loadPyodide === 'undefined') {
    let script = document.createElement('script');
    script.src = 'https://cdn.jsdelivr.net/pyodide/v0.21.3/full/pyodide.js';
    script.onload = async () => {
        let pyodide = await loadPyodide();
        window.pyodide = pyodide;
        console.log('loading pyodide.runPythonAsync(...). See in the following lines for the code');
        console.log('-----------------------  START PYTHON CODE  -------------------------------');
        console.log(`# python replace marker`);
        console.log('-----------------------  END PYTHON CODE    -------------------------------');
        pyodide.runPythonAsync(`# python replace marker`);
    };
    document.body.append(script)
}
"""



from typing import Iterator

from pynanto.webserver import Webserver


class AvailableWebservers:
    def __init__(self):
        self.classes = _webservers_classes()
        self.ids = list(map(lambda w: w.__name__, _webservers_classes()))

    def new_instance(self) -> Webserver:
        return self.classes[0]()

    def instances(self) -> Iterator[Webserver]:
        for webserver_class in self.classes:
            yield webserver_class()


def _webservers_classes():
    result = []
    try:
        from pynanto.webservers.fastapi import WsFastapi
        result.append(WsFastapi)
    except:
        pass
    try:
        from pynanto.webservers.flask import WsFlask
        result.append(WsFlask)
    except:
        pass
    try:
        from pynanto.webservers.tornado import WsTornado
        result.append(WsTornado)
    except:
        pass

    from pynanto.webservers.python_embedded import WsPythonEmbedded
    result.append(WsPythonEmbedded)

    return result


_available_webservers = None


def available_webservers() -> AvailableWebservers:
    global _available_webservers
    if _available_webservers is None:
        _available_webservers = AvailableWebservers()
    return _available_webservers



from typing import Callable, Awaitable

from js import HTMLElement, console
from pynanto.remote.asyncjs import waitAnimationFrame

from pynanto.remote.widget import Widget
import contextlib

AwaitableCallback = Callable[[], Awaitable[any]]


class SpinnerWidget(Widget):
    def __init__(self):
        super().__init__(html)
        self.div_loader: HTMLElement = self
        self.counter = 0

    async def spinnerSuspend(self, function: AwaitableCallback):
        await self.show_spinner()
        console.log(f"after increment {self.counter}")
        await waitAnimationFrame()
        try:
            await function()
        finally:
            await self.hide_spinner()

    async def hide_spinner(self):
        self.counter -= 1
        console.log(f"after increment {self.counter}")
        if self.counter == 0:
            await self.visible(False)

    async def show_spinner(self):
        if self.counter == 0:
            await self.visible(True)
        self.counter += 1

    @contextlib.asynccontextmanager
    async def context(self):
        await self.show_spinner()
        try:
            yield
        finally:
            await self.hide_spinner()

    async def visible(self, want_visible: bool):
        self.div_loader.style.display = '' if want_visible else 'none'
        await waitAnimationFrame()


# language=html
html = """<style>
    .spinner_base {
        border: 4px solid #f3f3f3; /* Light grey */
        border-top: 4px solid #3498db; /* Blue */
        border-radius: 50%;
        position: fixed;
        width: 30px;
        height: 30px;
        animation: spin 0.5s linear infinite;
    }

    #div_loader2 {
        left: 0px;
        top: 0px;
    }

    #div_loader {
        bottom: 0;
        right: 0;
    }

    @keyframes spin {
        0% {
            transform: rotate(0deg);
        }
        100% {
            transform: rotate(360deg);
        }
    }
</style>
<div id="div_loader" class="spinner_base" style="display: none"></div>

    
"""



import logging
import os
from dataclasses import dataclass
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path


def getLoggerFor(file):
    return logging.getLogger(os.path.basename(file))


logger = getLoggerFor(__file__)


@dataclass(frozen=True)
class LogConfig:
    when: str = 'midnight'
    interval: int = 1
    backup_count: int = 90
    format: str = "%(asctime)s [%(threadName)-12.12s] %(levelname)-5.5s %(name)s - %(message)s"


default = LogConfig()


def setup_logging(config: LogConfig = default):
    log_folder = (Path(__file__).parent.parent / 'data/logs').absolute()
    log_folder.mkdir(exist_ok=True, parents=True)
    logging.basicConfig(
        level=logging.INFO,
        format=config.format,
        handlers=[TimedRotatingFileHandler(
            log_folder / "app-log",
            when=config.when,
            interval=config.interval,
            backupCount=config.backup_count),
            logging.StreamHandler()
        ]
    )
    logger.info(f'Starting logging in folder {log_folder}')
    logger.info(f'Working directory: {os.getcwd()}')



from js import document
from pynanto.remote.components.ForwardTo import forward_to


class PnText:

    def __init__(self, text=None):
        div = document.createElement("div")
        self._proxy = div
        self.original_text = text
        if text is not None:
            div.innerHTML = text
        self._proxy = div
        document.body.append(div)

    def reset_text(self):
        self._proxy.innerHTML = self.original_text


forward_to(PnText, '_proxy')



from pynanto.unasync import unasync


def test_unsync():
    @unasync
    async def fun():
        return 'ok'

    assert fun() == 'ok'


def test_unsync_with_args():
    @unasync
    async def fun(arg1):
        return arg1

    assert fun('foo') == 'foo'



import asyncio
import sys
import traceback
from inspect import iscoroutinefunction
from threading import Thread


def unasync(f):
    if not iscoroutinefunction(f):
        return f
    result = None
    exc_info = None
    exception = None

    async def main_safe(*args, **kwargs):
        nonlocal result, exc_info, exception
        result = None
        exc_info = None
        exception = None
        try:
            result = await f(*args, **kwargs)
        except Exception as ex:
            exception = ex
            exc_info = sys.exc_info()
        print('done ')

    def start(*args, **kwargs):
        asyncio.run(main_safe(*args, **kwargs))

    def wrapper(*args, **kwargs):
        nonlocal result, exc_info, exception
        thread = Thread(target=start, args=args, kwargs=kwargs)
        thread.start()
        thread.join()
        if exc_info is not None:
            exc_type, exc_value, exc_traceback = exc_info
            print('', file=sys.stderr)
            traceback.print_exception(exc_type, exc_value, exc_traceback, file=sys.stderr)
            raise exception
        return result

    return wrapper



from functools import partial
from http import HTTPStatus
from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
from threading import Thread
from typing import Optional, Callable, Dict
from urllib.parse import urlparse, parse_qs

from pynanto.response import Response, Request
from pynanto.routes import Route
from pynanto.webserver import Webserver
from pynanto.webserver import wait_forever


class WsPythonEmbedded(Webserver):
    def __init__(self):
        super().__init__()
        self.thread: Optional[Thread] = None
        self._routes: Dict[str, Route] = {}

    def _setup_route(self, route: Route):
        self._routes[route.path] = route

    def _start_listen(self):
        httpd = ThreadingHTTPServer((self.host, self.port), partial(RequestHandler, handler=self._handler))

        def run():
            print(f'Starting embedded python web server on:\n'
                  f' - http://{self.host}:{self.port}\n'
                  f' - {self.localhost_url()}\n')
            httpd.serve_forever()

        self.thread = Thread(target=run, daemon=True)
        self.thread.start()

    def _handler(self, request: 'RequestHandler') -> bool:
        params, rpath = request.decode_request()
        route = self._routes.get(rpath, None)
        if route is None:
            nf = HTTPStatus.NOT_FOUND
            request.send_bytes(bytes(nf.phrase, 'utf8'), code=nf.value)
        else:
            req = Request(request.command, request.get_body(), request.get_content_type())
            resp = route.callback(req)
            content = resp.content
            if isinstance(content, str):
                content = bytes(content, 'utf8')
            if not isinstance(content, bytes):
                raise Exception(f'type of the content not supported: {type(content)}')
            request.send_bytes(content, content_type=resp.content_type)
        return True


class RequestHandler(SimpleHTTPRequestHandler):
    query_index = 4
    path_index = 2

    def __init__(self, *args
                 , handler: Callable[['RequestHandler'], bool]
                 , **kwargs):
        self.handler = handler
        super(RequestHandler, self).__init__(*args, **kwargs)

    def do_GET(self):
        self.handler(self)

    def do_POST(self):
        self.handler(self)

    def get_content_type(self) -> str:
        return self.headers.get('Content-Type')

    def get_body(self):
        if self.command != 'POST':
            return None
        content_len = int(self.headers.get('Content-Length'))
        body = self.rfile.read(content_len)
        return body

    def decode_request(self):
        p = urlparse(self.path)
        query_str = p[self.query_index]
        rpath = p[self.path_index]
        di = parse_qs(query_str)
        params = {k: v[0] for k, v in di.items()}
        return params, rpath

    def send_bytes(self, content, code=200, content_type='text/plain'):
        self.protocol_version = "HTTP/1.1"
        self.send_response(code)
        self.send_header("Content-Type", content_type)
        self.send_header("Content-Length", str(len(content)))
        self.send_header("Access-Control-Allow-Headers", "*")
        self.send_header("Access-Control-Allow-Methods", "*")
        self.send_header("Access-Control-Allow-Origin", "*")
        self.end_headers()
        self.wfile.write(content)

    def serve_file(self, directory, filename):
        self.directory = directory
        self.path = '/' + filename
        super(RequestHandler, self).do_GET()


if __name__ == '__main__':
    s = WsPythonEmbedded()
    s._setup_route(Route('/', lambda: Response('ciao', 'text/html')))
    s.start_listen()
    wait_forever()



from abc import ABC, abstractmethod
from typing import List, Iterable, Callable, Iterator

from pynanto.common.bundle import build_archive
from pynanto.resource import Resource

ResourceGenerator = Callable[[], Iterator[Resource]]


class Bundle(ABC):

    @abstractmethod
    def list(self) -> Iterable[Resource]:
        pass


class WatchableBundle(Bundle, ABC):
    @abstractmethod
    def is_changed(self):
        pass


def flatten_bundles(bundle_list: List[Bundle]) -> Iterable[Resource]:
    for bundle in bundle_list:
        yield from bundle.list()


class LambdaBundle(Bundle):
    def __init__(self, resource_generator: ResourceGenerator):
        self.resource_generator = resource_generator

    def list(self) -> Iterable[Resource]:
        gen = self.resource_generator()
        yield from gen


def strategy_no_cache(bundles: List[Bundle]) -> bytes:
    return build_archive(flatten_bundles(bundles))



import unittest
from datetime import date

from .time import first_day_next_month


class MyTestCase(unittest.TestCase):

    def test_1(self):
        target = first_day_next_month(today=date(2023, 1, 4))
        self.assertEqual(target, date(2023, 2, 1))

    def test_2(self):
        target = first_day_next_month(today=date(2022, 12, 5))
        self.assertEqual(target, date(2023, 1, 1))

    def test_3(self):
        target = first_day_next_month(today=date(2023, 1, 1))
        self.assertEqual(target, date(2023, 2, 1))

    def test_4(self):
        target = first_day_next_month(today=date(2023, 2, 1))
        self.assertEqual(target, date(2023, 3, 1))

    def test_5(self):
        target = first_day_next_month(today=date(2023, 2, 28))
        self.assertEqual(target, date(2023, 3, 1))


if __name__ == '__main__':
    unittest.main()



from typing import Callable, Optional, Dict

from pyodide.ffi import create_proxy, to_js

import js
from js import console, window, EventTarget

KeyboardEvent = object
_HotkeyHandler = Callable[[KeyboardEvent], Optional[bool]]


class Hotkey:
    def __init__(self, element: EventTarget):
        self.handlers: Dict[str, _HotkeyHandler] = dict()
        self.enable_log = False
        element.addEventListener('keydown', create_proxy(self._detect_hotkey), False)

    def add(self, hotkey: str, handler: _HotkeyHandler):
        self.handlers[hotkey] = handler

    @classmethod
    def keyboard_event(cls, e):
        return js.eval('(e) => e instanceof KeyboardEvent ')(e)

    def _detect_hotkey(self, e):
        if not self.keyboard_event(e):
            return
        key = ''
        if e.ctrlKey: key += 'CTRL-'
        if e.shiftKey: key += 'SHIFT-'
        if e.altKey: key += 'ALT-'
        if e.metaKey: key += 'META-'

        upc: str = e.key
        if len(upc) == 1: upc = upc.upper()

        key += upc
        if self.enable_log: console.log(key, to_js(e))
        handle = self.handlers.get(key, None)
        if handle is None:
            return

        res = handle(e)
        if not res:
            return

        console.log(f'prevent default for {key}')
        e.preventDefault()
        e.stopPropagation()


HotkeyWindow = Hotkey(window)



import os
import sys
import unittest
from pathlib import Path
from unittest import TestResult

from js import console


def unittest_main_fixed(module) -> TestResult:
    if isinstance(module, str):
        module = sys.modules[module]
    test_suit = unittest.TestLoader().loadTestsFromModule(module)

    result = _test_runner_run(test_suit)
    attach_result_meta(result)
    return result


def attach_result_meta(test_result: TestResult):
    from js import document
    meta = document.createElement('meta')
    meta.name = 'unittest-result'
    meta.content = f'wasSuccessful={test_result.wasSuccessful()}'
    document.head.append(meta)


def run_all_tests() -> TestResult:
    script_dir = os.path.dirname(__file__)
    # async tests seems to not be supported
    # there are async tests in `common`
    # so only test suite from  `browser` package
    browser = (Path(script_dir) / '..').absolute()

    loader = unittest.TestLoader()
    test_suit = loader.discover(str(browser))
    return _test_runner_run(test_suit)


def _test_runner_run(test_suit) -> TestResult:
    txt = 'test_result.txt'
    with open(txt, 'w') as f:
        res = unittest.TextTestRunner(stream=f).run(test_suit)
    print(f'TextTestRunner: {res}')
    c = Path(txt).read_text()
    console.log(c)
    return res



import asyncio
from threading import Thread
from typing import Optional, Awaitable

import tornado
import tornado.web

from pynanto import Webserver
from pynanto.routes import Route


class WsTornado(Webserver):
    def __init__(self):
        super().__init__()
        self.app = tornado.web.Application()
        self.thread: Optional[Thread] = None

    def _setup_route(self, route: Route):
        self.app.add_handlers(r".*", [(route.path, TornadoHandler, dict(route=route))])

    def _start_listen(self):
        async def run_async():
            self.app.listen(self.port, self.host)
            await asyncio.Event().wait()

        def run():
            asyncio.run(run_async())

        self.thread = Thread(target=run, daemon=True)
        self.thread.start()


class TornadoHandler(tornado.web.RequestHandler):

    def __init__(self, *args, **kwargs):
        self.route: Route = None
        super().__init__(*args, **kwargs)

    def initialize(self, route: Route) -> None:
        self.route = route

    def set_default_headers(self):
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Headers', '*')
        self.set_header('Access-Control-Allow-Methods', '*')

    def options(self):
        """/OPTIONS handler for preflight CORS checks."""
        self.set_status(204)
        self.finish()

    def get(self) -> None:
        response = self.route.callback()
        self.set_header("Content-Type", response.content_type)
        self.write(response.content)

    def post(self) -> None:
        raise_exception(self)

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        raise_exception(self)


def raise_exception(self: TornadoHandler):
    print("=" * 60)
    print(f'exc! {self.route}')
    raise Exception(self)
